# -*- coding: utf-8 -*-

# Run this app with `python visualization.py` and
# visit http://127.0.0.1:8050/ in your web browser.

from datetime import datetime as dt, timedelta as td
import pandas as pd

import dash
import dash_core_components as dcc
import dash_html_components as html
import plotly.express as px
from dash.dependencies import Input, Output, MATCH, State, ALL
import dash_table
import plotly.graph_objects as go

import os,sys,inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.append(current_dir)
import ui_text as t

parent_dir = os.path.dirname(current_dir)
sys.path.append(parent_dir)
from data_loader import es_premium_df, role_ids, translate_freq, segments, schools_names, date_column, conn, non_assistent_teachers, file_folder, users_df, vacation_info

sys.path.append(os.path.dirname(parent_dir))
from Modules.schools_helpers import filter_schools
from Modules.cohort_tiers import prepare_usage_line_for_plotting, cloe_data_count_in_timeframe, usage_in_window, merge_plotable_dfs
from Modules.time_series_helpers import TimeSeriesSampler

def register_callbacks(app):
    @app.callback(
        Output('role_filters', 'children'),
        [
            Input('role_select', 'value')
        ]
    )
    def update_user_filter(role):
        components = []

        if role_ids[role] == 2:
            return components + [dcc.Checklist(
                id={
                    'type': 'role_filters_checklist',
                    'index': 0
                },
                options=[
                    {'label': t.teachers_filter_text, 'value': 'non_assistent'},
                ],
                value=[],
                labelStyle={'display': 'inline-block'}
            )]
        
        return components

    flatten = lambda l: [item for sublist in l for item in sublist]

    @app.callback(
        Output('usage_lines_selector', 'options'),
        [
            Input('school_select', 'value'), 
            Input('role_select', 'value'),
            Input({'type':'role_filters_checklist', "index": ALL}, 'value')
        ]
    )
    def update_lines_select(school, role, assistent_teachers_switch):
        role_filter = (users_df.role_id == role_ids[role])
        school_filter = (users_df.school_id.isin(schools_names[school]))
        id_to_name_map = users_df[role_filter & school_filter]
        assistent_teachers_switch = flatten(assistent_teachers_switch)
        if 'non_assistent' in assistent_teachers_switch and role_ids[role] == 2:
            id_to_name_map = id_to_name_map[id_to_name_map.index.isin(non_assistent_teachers)]
        id_to_name_map = id_to_name_map[["full_name"]].drop_duplicates().copy()
        id_to_name_map = id_to_name_map.reset_index()                     
        id_to_name_map.columns = ["value", "label"]     
        id_to_name_map = id_to_name_map.to_dict("records")
        id_to_name_map = [{"value": 0, "label": "Geral"}] + id_to_name_map
        return id_to_name_map

    @app.callback(
        Output('engagement_chart', 'children'),
        [
            Input('school_select', 'value'), 
            Input('periodicity_select', 'value'), 
            Input('role_select', 'value'),
            Input('segment_select', 'value'),
            Input('date-picker-range', 'start_date'),
            Input('date-picker-range', 'end_date'),
            Input('event_select', 'value'),
            Input('usage_lines_selector', 'value'),
            Input({'type':'role_filters_checklist', "index": ALL}, 'value')
        ]
    )
    def update_graph(school, freq, role, segment_names, start_date, end_date, events, usage_lines, role_filters):
        role_filters = flatten(role_filters)
        segment_dfs = list(map(lambda x: segments[x].copy(), segment_names))
        if len(segment_dfs) == 0:
            return [
                html.H2(children=t.no_segment_header_msg, style={'textAlign': 'center', 'color': '#FF0000'}),
                html.P(children=t.no_segment_body_msg, style={'textAlign': 'center', 'color': '#FF0000'}),
            ]
        segment_df = pd.concat(segment_dfs)
        es_events = es_premium_df[es_premium_df.event_type.isin(events)]

        if 'non_assistent' in role_filters and role_ids[role] == 2:
            segment_df = segment_df[segment_df.index.isin(non_assistent_teachers)]
            es_events = es_events[es_events.user_id.isin(non_assistent_teachers)]

        if len(segment_df) == 0:
            return [
                html.H2(children=t.no_data_found_msg, style={'textAlign': 'center', 'color': '#FF0000'}),
                html.P(children=t.school_has_no_users_for_segment(role), style={'textAlign': 'center', 'color': '#FF0000'}),
            ]
        
        schools = schools_names[school]
        role_id = role_ids[role]    
        freq_txt = translate_freq[freq]

        ts_sampler = TimeSeriesSampler(freq_txt)
        date_index = pd.date_range(start=start_date, end=end_date, freq=ts_sampler.offset_alias)

        usage_dfs = []

        school_filter = (segment_df.school_id.isin(schools))
        role_filter = (segment_df.role_id == role_id)
        tos_filter = (segment_df.accepted_tos == True)

        registered_in_cloe = segment_df[school_filter & role_filter].copy()
        registered_in_cloe = cloe_data_count_in_timeframe(registered_in_cloe, date_index, ts_sampler, date_column, t.registered_in_cloe_curve_label)
        usage_dfs.append(registered_in_cloe)

        activated_1 = segment_df[school_filter & role_filter & tos_filter].copy()
        activated_1 = cloe_data_count_in_timeframe(activated_1, date_index, ts_sampler, date_column, t.accepted_tos_curve_label)
        usage_dfs.append(activated_1)

        individual_user_lines = 0
        if 0 in usage_lines:
            df = usage_in_window(schools, es_events, ts_sampler, role_id=role_id, users=segment_df)
            line_text = t.usage_curve_label(freq, "todos os usuários")
            plt_df = prepare_usage_line_for_plotting(df, date_column, date_index, line_text)
            usage_lines.remove(0)
            individual_user_lines += 1
            usage_dfs.append(plt_df)

        for user in usage_lines:
            es_events = es_premium_df[es_premium_df.user_id == user]
            df = usage_in_window(schools, es_events, ts_sampler, role_id=role_id, users=segment_df)
            line_text = t.usage_curve_label(freq, users_df[users_df.index == user].full_name.iloc[0])
            plt_df = prepare_usage_line_for_plotting(df, date_column, date_index, line_text)
            individual_user_lines += 1
            usage_dfs.append(plt_df)

        df = merge_plotable_dfs(usage_dfs, date_column)

        if df is None:
            return [
                html.H2(children=t.no_data_found_msg, style={'textAlign': 'center', 'color': '#FF0000'}),
                html.P(children=t.school_has_no_users_for_segment(role), style={'textAlign': 'center', 'color': '#FF0000'}),
            ]

        value_col_name = f'contagem'
        cols = [f'período', value_col_name, 'curvas']
        if df is None:
            df = pd.DataFrame(pd.DataFrame(columns=["date_month_year_day", "value",  "variables"]))
        df.columns = cols
        
        fig = px.line(df, x="período", y=value_col_name, color="curvas")

        for d in fig.data:
            d.update(mode='markers+lines')    
        
        if len(schools) == 1:
            create_vacation_shading(df[["período", value_col_name]], fig, ts_sampler, schools)

        fig.update_layout(legend=dict(
            orientation="v",
            yanchor="top",
            y=1.1 + 0.075 * len(usage_dfs),
            xanchor="left",
            x=0.01,
            borderwidth=1, 
            # itemsizing='constant',
            font=dict(
                family="Courier",
                size=12,
                color="black"
            ),
        ))

        return dcc.Graph(
            figure=fig
        )

    def get_week_start_date(d_str):
        fmt = "%Y-%m-%d"
        d = dt.strptime(d_str,fmt)
        return (d - td(days=d.weekday())).strftime(fmt)

    def get_one_week_ahead(d_str):
        fmt = "%Y-%m-%d"
        d = dt.strptime(d_str,fmt)
        return (d + td(days=7)).strftime(fmt)

    def get_time_bounds(start, end):
        original_end_date = end
        start, end = get_week_start_date(start), get_week_start_date(end)
        if start == end or not end == original_end_date:
            end = get_one_week_ahead(end)
        return start, end

    def create_vacation_shading(df, fig, ts_sampler, schools):
        showlegend=True
        for d in vacation_info[vacation_info.id.isin(schools)].to_dict("records"):
            start, end = get_time_bounds(d["init"], d["end"])
            x = pd.date_range(start=start, end=end, freq=ts_sampler.offset_alias).strftime("%Y/%m/%d").to_list()
            y = df.groupby(by="período").max()
            y = y[y.index.isin(x)].contagem
            area = go.Scatter(x=x, y=y, 
                fill='tozeroy',
                fillcolor='rgba(255, 125, 125, 0.3)',
                mode='none',
                name=t.chart_vacation_area_label,
                showlegend=showlegend,
            )
            showlegend=False
            fig.add_trace(area)

    @app.callback(Output('tabs-content', 'children'),
        [
            Input('tabs', 'value'),
            Input('school_select', 'value'), 
            Input('periodicity_select', 'value'), 
            Input('role_select', 'value'),
            Input('segment_select', 'value'),
            Input('date-picker-range', 'start_date'),
            Input('date-picker-range', 'end_date'),
            Input('event_select', 'value')
        ]
    )
    def on_reload_tab(*args):
        tab = args[0]
        if tab == 'tab-1':
            return dcc.Loading(id = 'engagement_chart', children=html.H2(children=t.cant_load_table_msg, style={'textAlign': 'center'}), type='circle')
        elif tab == 'tab-2':
            return dcc.Loading(id = 'table_div', children=html.H2(children=t.cant_load_table_msg, style={'textAlign': 'center'}), type='circle')

    @app.callback(Output('table_div', 'children'),
        [
            Input('school_select', 'value'), 
            Input('role_select', 'value'),
            Input('segment_select', 'value'),
            Input('date-picker-range', 'start_date'),
            Input('date-picker-range', 'end_date')
        ]
    )
    def on_data_set_table(school, role, segment_names, start_date, end_date):
        df = filter_schools(schools_names[school], es_premium_df)
        df.date = pd.to_datetime(df.date, utc=True).dt.tz_convert('America/Sao_Paulo')
        df = df.set_index("date")
        df = df.loc[start_date:end_date].reset_index()
        filter_role = (df["user_properties.role_id"] == role_ids[role])
        df = df[filter_role]
        segment_dfs = list(map(lambda x: segments[x].copy(), segment_names))
        segment_df = pd.concat(segment_dfs).reset_index()
        df = df[df.user_id.isin(segment_df.user_id)]
        if len(df) == 0:
            return [
                html.H2(children=t.no_data_found_msg, style={'textAlign': 'center', 'color': '#FF0000'}),
                html.P(children=t.please_change_parameters_msg, style={'textAlign': 'center', 'color': '#FF0000'}),
            ]

        user_names = users_df[["full_name"]].drop_duplicates(subset=["full_name"])

        df = df.set_index("user_id").join(user_names).reset_index()

        df = df[["date", "full_name", "user_id"]]

        df = df.drop_duplicates(subset=["user_id"], keep="last")
        df = df.dropna()
        df.columns = ["último evento", "usuário", "id no banco da cloe"]

        return html.Div([
                dash_table.DataTable(
                    id='detailed-table',
                    columns=[{'name': i, 'id': i} for i in df.columns],
                    data=df.to_dict("rows")
                )
            ])